
		/*********************************************************
		** This program finds a binary vessels map.  It reads
		** the MSF (matched spatial filter) image, to save time.
		** This image is thresholded at the highest threshold
		** which gives 15000 pixels.  These pixels are then
		** thinned and broken at all branches, so that each
		** piece contains two endpoints only.  Small pieces
		** are discarded.  The remaining endpoints initialize
		** a grow queue.
		**
		** For each endpoint in the grow queue, a segment is
		** probed with a local monotonically decreasing threshold.
		** The local threshold decreases so long as
		**  (a) the segment does not grow too big
		**  (b) the threshold does not bottom out 
		**  (c) the segment does not touch >1 other segments
		**  (d) the segment does not fringe upon other segments
		**  (e) the segment does not grow a loop
		**  (f) the ratio of size:branches for the segment stays high
		** Upon exiting, the grown segment is kept if its last
		** usable size (size before exit condition reached) is
		** above a minimum.  If it is kept, then its endpoints
		** are added to the grow queue.  The process ends when
		** when the grow queue is empty.
		**
		** This is version grow-network.6.c ported from unix.
		*********************************************************/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

#define GRAPHICS_ON		0				/* 1 for partial animation, 2 for full animation */
#define START_PTS_ON	0				/* 1 to display starting point for each segment before it grows */
#define REASONS_ON		0				/* 1 to display text reason for halting each segment growth */

#define MIN_SEGMENT_SIZE		10		/* units are pixels */
#define THRESH_COUNT			15000 /* 35000	/* (was 15000) units are pixels */

#define MAX_PROBE_QUEUE	500000
#define MAX_QUEUE		500000

#define MIN_EVALUATE_SIZE		30 		/* units are pixels */
#define MIN_PROBE_SIZE			150 /* 60		/* (was 150) units are pixels */
#define MAX_PROBE_SIZE			3500 /* 2000	/* (was 3500) units are pixels */
#define MAX_FRINGE_PER_SIZE		0.34 	/* units are percentage */
#define MAX_FRINGE_PER_TOUCH	99999.9	/* 20.0	/* units are pixels/touch */
#define MIN_SIZE_PER_BRANCH		150.0 	/* units are pixels/branch */
#define MAX_CURVATURE			99999.9	/* 1.8		/* units are dimensionless */


int MinProbeSize[10]={180,160,140,120,100,80,60,40,20,0};
int MaxProbeSize[10]={1000,2000,3000,4000,5000,6000,7000,8000,9000,10000};
double MaxFringePerSize[10]={.1,.14,.18,.22,.26,.3,.34,.38,.42,.46};
double MinSizePerBranch[10]={300,275,250,225,200,175,150,125,100,75};
int ThreshCount[10]={5000,10000,15000,20000,25000,30000,35000,40000,45000,50000};
//int ThreshCount[10]={500,1000,1500,2000,2500,3000,3500,4000,4500,5000}; // for Laser images guy I was helping


void ProbeVesselNetwork(unsigned char *MSFImage,
						int ROWS,int COLS,
						unsigned char *ThreshImage)

{
HDC				hDC;
char			text[100];
unsigned char	*ThinImage,*GraphicsImage;
int				a,i,r,c;
int				hist[256],count,thresh;

int		sy,sx,vy,vx,Label,*indices;
int		*endpoints,total_ends,*branches,total_branches;
int		Alternate,MinProbeIntensity,StartProbeIntensity,LastSize;
int		touch_indices[256],total_touches,fringe,StopReason;

int		*Queue;	  /* for PaintFill(), allocate here, once only */
int		QueueCount,*ProbeQueue,qh,qt;
double	curvature;


if ((GraphicsImage=(unsigned char *)calloc(3,ROWS*COLS)) == NULL  ||
	(ThinImage=(unsigned char *)calloc(1,ROWS*COLS)) == NULL  ||
	(indices=(int *)calloc(ROWS*COLS,sizeof(int))) == NULL  ||
	(Queue=(int *)calloc(MAX_QUEUE,sizeof(int))) == NULL  ||
	(ProbeQueue=(int *)calloc(MAX_PROBE_QUEUE,sizeof(int))) == NULL  ||
	(endpoints=(int *)calloc(MAX_PROBE_QUEUE,sizeof(int))) == NULL  ||
	(branches=(int *)calloc(MAX_PROBE_QUEUE,sizeof(int))) == NULL)
  {
  MessageBox(NULL,"Error","Unable to allocate memory for images.",MB_OK | MB_APPLMODAL);
  return;
  }

		/*********************************************************
		** Unlabel 4-deep border of MSF image (from circle).
		** A hack to take care of the MSF response to the border.
		*********************************************************/

for (r=0; r<ROWS; r++)
  {
  c=0;
  while (c < COLS  &&  MSFImage[r*COLS+c] == 0)
    c++;
  if (c < COLS-4)
    {
    i=0;
    while (i < 4)
      { MSFImage[r*COLS+c+i]=0; i++; }
    }
  c=COLS-1;
  while (c >= 0  &&  MSFImage[r*COLS+c] == 0)
    c--;
  if (c >= 4)
    {
    i=0;
    while (i < 4)
      { MSFImage[r*COLS+c-i]=0; i++; }
    }
  }
for (c=0; c<COLS; c++)
  {
  r=0;
  while (r < ROWS  &&  MSFImage[r*COLS+c] == 0)
    r++;
  if (r < ROWS-4)
    {
    i=0;
    while (i < 4)
      { MSFImage[(r+i)*COLS+c]=0; i++; }
    }
  r=ROWS-1;
  while (r >= 0  &&  MSFImage[r*COLS+c] == 0)
    r--;
  if (r >= 4)
    {
    i=0;
    while (i < 4)
      { MSFImage[(r-i)*COLS+c]=0; i++; }
    }
  }
	/* probably don't want this code anymore */

		/*********************************************************
		** Make histogram of MSF image for thresholding,
		** then threshold.  Index thresholded pixels for thinning.
		*********************************************************/

for (i=0; i<256; i++)
  hist[i]=0;
for (i=0; i<ROWS*COLS; i++)
  hist[MSFImage[i]]++;
count=hist[255];
thresh=255;
while (count < ThreshCount[VesselsParameters]  &&  thresh > 0)
  {
  thresh--;
  count+=hist[thresh];
  }
if (thresh == 0)
  {
  MessageBox(NULL,"Error","THRESH_COUNT pixels not found",MB_OK | MB_APPLMODAL);
  return;
  }
count=0;
for (i=0; i<ROWS*COLS; i++)
  if (MSFImage[i] >= thresh)
    {
    GraphicsImage[i*3+0]=GraphicsImage[i*3+1]=GraphicsImage[i*3+2]=255;
    ThreshImage[i]=255;
    indices[count]=i;
    count++;
    }
  else
    {
    GraphicsImage[i*3+0]=GraphicsImage[i*3+1]=GraphicsImage[i*3+2]=0;
    ThreshImage[i]=0;
    }

		/*********************************************************
		** Display thresh image
		*********************************************************/

if (GRAPHICS_ON)
  {
  MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
  PaintImage();
  MessageBox(NULL,"Thresholded MSF","Probing...",MB_OK | MB_APPLMODAL);
  }

		/*********************************************************
		** Thin the image.  Unlabel all branchpoints.  Relabel
		** the image so that each connected component has a
		** different label, so that sizes can be determined.
		** Discard small segments.  Intialize grow-queue with
		** maximum MSF pixel for each remaining segment.
		*********************************************************/

ThinPixels(ThreshImage,ROWS,COLS,indices,count,255);
BranchesAndEnds(ThreshImage,ROWS,COLS,indices,count,endpoints,&total_ends,branches,&total_branches,MaxProbeSize[VesselsParameters],255);
for (i=0; i<total_branches; i++)
  {
  for (r=branches[i]/COLS-1; r<=branches[i]/COLS+1; r++)
    for (c=branches[i]%COLS-1; c<=branches[i]%COLS+1; c++)
      if (r >= 0  &&  r < ROWS  &&  c >= 0  &&  c < COLS)
        ThreshImage[r*COLS+c]=0;
  }
Label=254;
QueueCount=0;
for (r=0; r<ROWS; r++)
  for (c=0; c<COLS; c++)
    {
    if (ThreshImage[r*COLS+c] != 255)
      continue;
    count=PaintFill(ThreshImage,ROWS,COLS,r,c,255,Label,Queue,MAX_QUEUE,indices);
    if (count < MIN_SEGMENT_SIZE)
      {		/* get rid of small segments */
      PaintFill(ThreshImage,ROWS,COLS,r,c,Label,0,Queue,MAX_QUEUE,indices);
      continue;
      }
    BranchesAndEnds(ThreshImage,ROWS,COLS,indices,count,endpoints,&total_ends,branches,&total_branches,MaxProbeSize[VesselsParameters],Label);
    if (QueueCount+total_ends > MAX_PROBE_QUEUE)
      {
	  MessageBox(NULL,"Error","MAX_PROBE_QUEUE exceeded!",MB_OK | MB_APPLMODAL);
	  return;
      }
    for (i=0; i<total_ends; i++)
      ProbeQueue[QueueCount+i]=endpoints[i];
    QueueCount+=total_ends;
    Label--;
    if (Label < 20)
      {
      for (i=0; i<ROWS*COLS; i++)
        if (ThreshImage[i] >= 19  &&  ThreshImage[i] < 255)
          ThreshImage[i]=254;
      Label=253;
      }
    }


		/*********************************************************
		** Display CC image
		*********************************************************/

if (GRAPHICS_ON)
  {
  for (i=0; i<ROWS*COLS; i++)
    if (ThreshImage[i])
      GraphicsImage[i*3+0]=GraphicsImage[i*3+1]=GraphicsImage[i*3+2]=255;
    else
      GraphicsImage[i*3+0]=GraphicsImage[i*3+1]=GraphicsImage[i*3+2]=0;
  for (i=0; i<QueueCount; i++)
	{
    sy=ProbeQueue[i]/COLS; sx=ProbeQueue[i]%COLS;
    for (r=sy-7; r<=sy+7; r++)
      for (c=sx-7; c<=sx+7; c++)
		{
        GraphicsImage[(r*COLS+c)*3+0]=255;
        GraphicsImage[(r*COLS+c)*3+1]=GraphicsImage[(r*COLS+c)*3+2]=0;
		}
	}
  MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
  PaintImage();
  sprintf(text,"%d points in queue",QueueCount);
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  MessageBox(NULL,"Initial queue points","Probing...",MB_OK | MB_APPLMODAL);
  }


		/*********************************************************
		** Determine curvature of each connected component.
		** Discard if component is highly curved.
		*********************************************************/

if(0){
	/* TBD */
}

		/*********************************************************
		** Locally grow from all points in ProbeQueue
		*********************************************************/

for (i=0; i<ROWS*COLS*3; i++)
  GraphicsImage[i]=0;
for (i=0; i<ROWS*COLS; i++)
  ThinImage[i]=ThreshImage[i]=0;
qh=QueueCount;	/* head of queue; next (empty) spot in line */
qt=0;			/* tail of queue; next (occupied) spot in line to process */
Label=254;

while (qt != qh)
  {

  sprintf(text,"%d points in queue",(qh > qt ? qh-qt : qh+MAX_QUEUE-qt));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);

  i=GetStartGrowPoint(ThinImage,ThreshImage,ROWS,COLS,ProbeQueue[qt],
		&sy,&sx,&vy,&vx);
  if (START_PTS_ON)
	{
    for (r=ProbeQueue[qt]/COLS-7; r<=ProbeQueue[qt]/COLS+7; r++)
	  for (c=ProbeQueue[qt]%COLS-7; c<=ProbeQueue[qt]%COLS+7; c++)
        {
        GraphicsImage[(r*COLS+c)*3+0]=255;
        GraphicsImage[(r*COLS+c)*3+1]=GraphicsImage[(r*COLS+c)*3+2]=0;
        }
	MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
	PaintImage();
	sprintf(text,"%d points in queue",(qh > qt ? qh-qt : qh+MAX_QUEUE-qt));
	hDC=GetDC(MainWnd);
    TextOut(hDC,0,0,text,strlen(text));
    ReleaseDC(MainWnd,hDC);
	}
  if (i == 0)
    {	/* can't figure out direction to grow endpoint */
	/* erase this branch? (code to be added here) */
	if (START_PTS_ON)
      MessageBox(NULL,"Couldn't find start point","Probing...",MB_OK | MB_APPLMODAL);
    qt=(qt+1)%MAX_PROBE_QUEUE;
    continue;
    }
  if (i == 2)
    ClipGrowPoint(ThreshImage,ROWS,COLS,sy,sx,vy,vx);
  /* printf("End %d,%d (probe %d) ",sy,sx,MSFImage[sy*COLS+sx]); fflush(stdout); */
  MinProbeIntensity=StartProbeIntensity=MSFImage[sy*COLS+sx];
  Alternate=0;
  StopReason=0;
  count=0;
  do
    {
    LastSize=count;	/* size of segment from previous probe */
    if (!ProbeRegion(MSFImage,ThreshImage,ROWS,COLS,sy,sx,
		MinProbeIntensity,Alternate+1,Queue,MAX_QUEUE,indices,&count,&fringe,
		touch_indices,&total_touches,MaxProbeSize[VesselsParameters]))
      break;	/* growing out of control (too big) */

	if (GRAPHICS_ON == 2  &&  count > LastSize)
	  {
	  for (a=0; a<count; a++)
		{
		GraphicsImage[indices[a]*3+0]=255;
		GraphicsImage[indices[a]*3+1]=GraphicsImage[indices[a]*3+2]=0;
		}
	  MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
	  PaintImage();
	  }

    Alternate=(Alternate+1)%2;
    MinProbeIntensity--;
    if (MinProbeIntensity <= 1)
      {
      StopReason=1;
      break;	/* probe not going anywhere */
      }

    if (count < MIN_EVALUATE_SIZE)
      {			/* grow at least until it gets big enough to evaluate */
      if (total_touches > 1)
        {
        StopReason=2;	/* unless it already connects two pieces */
        break;
        }
      continue;
      }
    if ((double)fringe/(double)count > MaxFringePerSize[VesselsParameters])
      {
      StopReason=3;
      break;	/* growing along border of pixels already labeled as vessel */
      }
    if (total_touches > 0  &&  (double)fringe/(double)total_touches > MAX_FRINGE_PER_TOUCH)
      {
      StopReason=4;
      break;	/* growing along border of pixels already labeled as vessel */
      }

    for (i=0; i<count; i++)	/* copy new growth, thin and evaluate it */
      ThinImage[indices[i]]=ThreshImage[indices[i]];
    ThinPixels(ThinImage,ROWS,COLS,indices,count,((Alternate+1)%2)+1);
    BranchesAndEnds(ThinImage,ROWS,COLS,indices,count,endpoints,&total_ends,branches,&total_branches,MaxProbeSize[VesselsParameters],((Alternate+1)%2)+1);

    if (0 /* total_branches+2 > total_ends  &&  0 */)
      {
      StopReason=5;
      break;	/* has a loop somewhere */
      }
    if ((total_branches > 1  &&  (double)count/(double)total_branches < MinSizePerBranch[VesselsParameters])
		  ||  total_branches > 10)
      {
      StopReason=6;
      break;	/* not enough substance for so many branches */
      }

    SegmentCurvature(ThinImage,ROWS,COLS,((Alternate+1)%2)+1,indices,count,
			endpoints,total_ends,branches,total_branches,Queue,MAX_QUEUE,&curvature);
    if (curvature > MAX_CURVATURE)
      {
      StopReason=7;	/* too curvy -- doesn't look like vessel */
      break;
      }

    if (total_touches >= 2)
      {
      StopReason=2;
      break;	/* connected multiple pieces, and everything else ok */
      }
    }
  while (count < MaxProbeSize[VesselsParameters]);
  MinProbeIntensity++;	/* puts it back to value used in last probe */

  if (GRAPHICS_ON)
    {
	for (a=0; a<count; a++)
	  {
	  GraphicsImage[indices[a]*3+0]=GraphicsImage[indices[a]*3+1]=
	  GraphicsImage[indices[a]*3+2]=0;
	  }
/*	MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
	PaintImage(); */
	}

  for (i=0; i<count; i++)
    ThinImage[indices[i]]=ThreshImage[indices[i]]=0;	/* erase growth */

  if (StopReason == 0)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Probe too big?","StopReason=0",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }
  if (StopReason == 1)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Probe didn't go anywhere","StopReason=1",MB_OK | MB_APPLMODAL);
    qt=(qt+1)%MAX_PROBE_QUEUE;
    continue;
    }
  if (StopReason == 2)
	if (REASONS_ON)
	  MessageBox(NULL,"Connected pieces","StopReason=2",MB_OK | MB_APPLMODAL);
  if (StopReason == 3)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Started to fringe, based on size","StopReason=3",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }
  if (StopReason == 4)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Started to fringe, based on # branches","StopReason=4",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }
  if (StopReason == 5)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Looped","StopReason=5",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }
  if (StopReason == 6)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Too many branches per size","StopReason=6",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }
  if (StopReason == 7)
    {
	if (REASONS_ON)
	  MessageBox(NULL,"Too curvy","StopReason=7",MB_OK | MB_APPLMODAL);
    MinProbeIntensity++;
    count=LastSize;
    if (MinProbeIntensity > StartProbeIntensity)
      {
      printf("\tProbe bottomed out\n");
      qt=(qt+1)%MAX_PROBE_QUEUE;
      continue;
      }
    }

  if (count < MinProbeSize[VesselsParameters]  &&  StopReason != 2)
    {
    ProbeRegion(MSFImage,ThreshImage,ROWS,COLS,sy,sx,
				MinProbeIntensity,3,Queue,MAX_QUEUE,indices,&count,&fringe,
				touch_indices,&total_touches,MaxProbeSize[VesselsParameters]);
    printf("\tInsufficient LastSize %d \n",count);
    qt=(qt+1)%MAX_PROBE_QUEUE;
	if (GRAPHICS_ON)
	  {
	  for (a=0; a<count; a++)
		{
		GraphicsImage[indices[a]*3+0]=GraphicsImage[indices[a]*3+1]=
		GraphicsImage[indices[a]*3+2]=100;
		}
	  MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
	  PaintImage();
	  }
    continue;
    }

  ProbeRegion(MSFImage,ThreshImage,ROWS,COLS,sy,sx,
				MinProbeIntensity,Label,
				Queue,MAX_QUEUE,indices,&count,&fringe,
				touch_indices,&total_touches,MaxProbeSize[VesselsParameters]);
  for (i=0; i<count; i++)	/* copy new growth, thin, and find ends */
    ThinImage[indices[i]]=ThreshImage[indices[i]];
  ThinPixels(ThinImage,ROWS,COLS,indices,count,Label);
  BranchesAndEnds(ThinImage,ROWS,COLS,indices,count,endpoints,&total_ends,branches,&total_branches,
				  MaxProbeSize[VesselsParameters],Label);
  /* printf("\t\tLabel %d (probe %d) is size %d b/e %d/%d SR=%d\n",Label,
		MinProbeIntensity,count,total_branches,total_ends,StopReason); */
  if (QueueCount+total_ends > MAX_PROBE_QUEUE)
    {
	MessageBox(NULL,"MAX_PROBE_QUEUE exceeded!","Error in ProbeNetwork()",
				MB_OK | MB_APPLMODAL);
    exit(0);
    }
  for (i=0; i<total_ends; i++)
    ProbeQueue[(qh+i)%MAX_PROBE_QUEUE]=endpoints[i];
  qh=(qh+total_ends)%MAX_PROBE_QUEUE;

  /* printf("%d new points, %d now in queue (head=%d tail=%d)\n",total_ends,
	(qh+MAX_PROBE_QUEUE-qt)%MAX_PROBE_QUEUE,qh,qt); */

  if (GRAPHICS_ON)
	{
	for (a=0; a<count; a++)
	  {
	  GraphicsImage[indices[a]*3+0]=GraphicsImage[indices[a]*3+1]=
	  GraphicsImage[indices[a]*3+2]=255;
	  }
	MakeDisplayImage(GraphicsImage,ROWS,COLS,3);
	PaintImage();
	}

  qt=(qt+1)%MAX_PROBE_QUEUE;
  Label--;
  if (Label < 10)
	{
	Label=254;
	/*
	MessageBox(NULL,"Labels exceeded in GrowNetwork()","Error",MB_OK | MB_APPLMODAL);
	break;
	*/
	}
  }

		/*********************************************************
		** Binarize image, eliminating artificial probe barriers.
		** Skip this code to debug, to see individual probe pieces.
		*********************************************************/
for (i=0; i<ROWS*COLS; i++)
  if (ThreshImage[i] >= 4)
	ThreshImage[i]=255;
  else
	ThreshImage[i]=0;

		/*********************************************************
		** Free up memory that was dynamically allocated.
		*********************************************************/

free(GraphicsImage);
free(ThinImage);
free(indices);
free(endpoints);
free(branches);
free(Queue);
free(ProbeQueue);
}




		/*********************************************************
		** When a local probe (threshold) starts, part of it can
		** grow backward along the sides of the segment the growth
		** is supposed to be extending.  To prevent this, a clipping
		** line (label=3) is added perpendicular to the segment
		** at the local probe point, before probing starts.
		** The clipping line, which is 2 8-connected pixels thick
		** and 8 pixels long, can be removed after probing.
		*********************************************************/

void ClipGrowPoint(unsigned char *FullImage,	/* image to which clipping line will be added */
		int ROWS,int COLS,
		int gy,int gx,		/* point to probe from */
		int vy,int vx)		/* vector back along center line of segment */

{
int	line_indices[20],count,k,lx,ly;

lx=gx; ly=gy;
MakeLineIndices(lx+vy,ly-vx,lx-vy,ly+vx,ROWS,COLS,line_indices,&count);
for (k=0; k<count; k++)
  if (FullImage[line_indices[k]] == 0  &&  !(line_indices[k] == ly*COLS+lx))
    FullImage[line_indices[k]]=3;
ly+=vy/4; if (!(abs(vy) == abs(vx))) lx+=vx/4;
MakeLineIndices(lx+vy,ly-vx,lx-vy,ly+vx,ROWS,COLS,line_indices,&count);
for (k=0; k<count; k++)
  if (FullImage[line_indices[k]] == 0  &&  !(line_indices[k] == ly*COLS+lx))
    FullImage[line_indices[k]]=3;
}




		/*********************************************************
		** When an image is thinned, the endpoints of segments are
		** slightly interior to the original boundaries.  In order
		** to probe endpoints locally, the thinned endpoint must be
		** extrapolated back to the corresponding boundary of the
		** pre-thinned (full) image.  This routine does that.
		**
		** If the endpoint is not yet labeled, this routine
		** assumes it is starting a new segment.  1 is returned.
		** If the endpoint is extending a segment, 2 is returned.
		** If a new endpoint cannot be found, 0 is returned.
		*********************************************************/

int GetStartGrowPoint(unsigned char *ThinImage,	/* endpoints are slightly shrunk */
		unsigned char	*FullImage,	/* before thinning */
		int		ROWS,int COLS,
		int		endpoint,	/* coded y*COLS+x */
		int		*gy,int *gx,	/* returned point */
		int		*vy,int *vx)	/* returned vector, back along segment */

{
int	i,r,c,r2,c2,dy,dx,count,line_indices[30],dx1,dy1;

if (ThinImage[endpoint] == 0)
  {	/* not extending previous segment -- starting new one */
  *gy=endpoint/COLS;
  *gx=endpoint%COLS;
  (*vx)=(*vy)=0;
  return(1);	/* just grow from here */
  }
	/* search for instances of same label at 4-pixel radius */
r=endpoint/COLS; c=endpoint%COLS;
count=0;
for (r2=-4; r2<=4; r2++)
  for (c2=-4; c2<=4; c2++)
    {
    if (!(abs(r2) == 4  ||  abs(c2) == 4))
      continue;		/* looking for pixel on 8x8 surrounding box */
    if (ThinImage[(r+r2)*COLS+c+c2] == ThinImage[r*COLS+c])
      {
      if (count == 1)
        { dx1=dx; dy1=dy; }
      dy=r2;
      dx=c2;
      count++;
      }
    }
if (count == 2  &&  abs(dy1-dy) <= 1  &&  abs(dx1-dx) <= 1)
  count=1;	/* thinning problem -- not quite 8-connected */
if (count > 1)
  {
/*
  printf("Endpoint %d,%d too close to branch\n",r,c);
*/
  return(0);
  }
if (count < 1)
  {
/*
  printf("Couldn't find any reference for endpoint %d %d (is loop?)\n",r,c);
*/
  return(0);
  }
MakeLineIndices(c,r,c-3*dx,r-3*dy,ROWS,COLS,line_indices,&count);
for (i=0; i<count; i++)
  if (FullImage[line_indices[i]] != FullImage[endpoint])
    break;
if (i == count)
  {
/*
  printf("Couldn't extend line out of segment for %d,%d\n",r,c);
*/
  return(0);
  }
*gy=line_indices[i]/COLS;
*gx=line_indices[i]%COLS;
*vy=dy; *vx=dx;
return(2);
}








	/*
	** Labels in LabelImage are used as follows:
	**	0 => untouched
	**	1 => probe label
	**	2 => alternate probe label (so don't have to undo probe)
	**	3 => grown, checked, labeled as non-interesting
	**	255 on down => labeled regions
	**
	** returns 1 if probe does not exceed MaxProbeSize; returns 0
	** if it does
	*/

int ProbeRegion(unsigned char	*TestImage,	/* image to test criteria within */
	unsigned char	*LabelImage,	/* image which contains labels during grow */
	int		ROWS,int COLS,	/* size of image */
	int		r,int c,	/* starting point to grow region */
	int		MinIntensity,	/* test criteria */
	int		TrackLabel,	/* index to write to LabelImage during probe */
	int		*queue,		/* chunk of memory to use during probe */
	int		max_queue,	/* size of chunk of memory */
	int		*indices,	/* output:  indices of points in region */
	int		*total_indices,	/* output:  count of pixels in region */
	int		*Contact,	/* output:  count of border-contact pixels */
	int		*touch_labels,	/* output:  labels of regions probe touches */
	int		*total_touches,	/* output:  count of regions touched */
	int		MaxProbeSize)	/* clip-size, to speed things up */

{
int	qh,qt,count,i;


*total_touches=0;
if (TestImage[r*COLS+c] < MinIntensity  ||  LabelImage[r*COLS+c] >= 3)
  {
  *total_indices=0;
  return(1);
  }
LabelImage[r*COLS+c]=TrackLabel;
queue[0]=r*COLS+c;
qh=1;
qt=0;
indices[0]=r*COLS+c;
count=1;
(*Contact)=0;
while (qt != qh  &&  count < MaxProbeSize)
  {
  if (queue[qt]/COLS > 0  &&  LabelImage[queue[qt]-COLS] != TrackLabel)
    {
    if (LabelImage[queue[qt]-COLS] >= 3)
      {
      (*Contact)++;
      for (i=0; i<(*total_touches); i++)
        if (touch_labels[i] == LabelImage[queue[qt]-COLS])
          break;
      if (i == (*total_touches)  &&  LabelImage[queue[qt]-COLS] != 3)
        {
        touch_labels[(*total_touches)]=LabelImage[queue[qt]-COLS];
        (*total_touches)++;
        }
      }
    else if (TestImage[queue[qt]-COLS] >= MinIntensity)
      {
      LabelImage[queue[qt]-COLS]=TrackLabel;
      indices[count]=queue[qt]-COLS;
      count++;
      queue[qh]=queue[qt]-COLS;
      qh=(qh+1)%max_queue;
      if (qh == qt  ||  count == max_queue)
        {
        MessageBox(NULL,"Error","MAX_QUEUE exceeded!",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
    }
  if (queue[qt]/COLS < ROWS-1  &&  LabelImage[queue[qt]+COLS] != TrackLabel)
    {
    if (LabelImage[queue[qt]+COLS] >= 3)
      {
      (*Contact)++;
      for (i=0; i<(*total_touches); i++)
        if (touch_labels[i] == LabelImage[queue[qt]+COLS])
          break;
      if (i == (*total_touches)  &&  LabelImage[queue[qt]+COLS] != 3)
        {
        touch_labels[(*total_touches)]=LabelImage[queue[qt]+COLS];
        (*total_touches)++;
        }
      }
    else if (TestImage[queue[qt]+COLS] >= MinIntensity)
      {
      LabelImage[queue[qt]+COLS]=TrackLabel;
      indices[count]=queue[qt]+COLS;
      count++;
      queue[qh]=queue[qt]+COLS;
      qh=(qh+1)%max_queue;
      if (qh == qt  ||  count == max_queue)
        {
        MessageBox(NULL,"Error","MAX_QUEUE exceeded!",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
    }
  if (queue[qt]%COLS > 0  &&  LabelImage[queue[qt]-1] != TrackLabel)
    {
    if (LabelImage[queue[qt]-1] >= 3)
      {
      (*Contact)++;
      for (i=0; i<(*total_touches); i++)
        if (touch_labels[i] == LabelImage[queue[qt]-1])
          break;
      if (i == (*total_touches)  &&  LabelImage[queue[qt]-1] != 3)
        {
        touch_labels[(*total_touches)]=LabelImage[queue[qt]-1];
        (*total_touches)++;
        }
      }
    else if (TestImage[queue[qt]-1] >= MinIntensity)
      {
      LabelImage[queue[qt]-1]=TrackLabel;
      indices[count]=queue[qt]-1;
      count++;
      queue[qh]=queue[qt]-1;
      qh=(qh+1)%max_queue;
      if (qh == qt  ||  count == max_queue)
        {
        MessageBox(NULL,"Error","MAX_QUEUE exceeded!",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
    }
  if (queue[qt]%COLS < COLS-1  &&  LabelImage[queue[qt]+1] != TrackLabel)
    {
    if (LabelImage[queue[qt]+1] >= 3)
      {
      (*Contact)++;
      for (i=0; i<(*total_touches); i++)
        if (touch_labels[i] == LabelImage[queue[qt]+1])
          break;
      if (i == (*total_touches)  &&  LabelImage[queue[qt]+1] != 3)
        {
        touch_labels[(*total_touches)]=LabelImage[queue[qt]+1];
        (*total_touches)++;
        }
      }
    else if (TestImage[queue[qt]+1] >= MinIntensity)
      {
      LabelImage[queue[qt]+1]=TrackLabel;
      indices[count]=queue[qt]+1;
      count++;
      queue[qh]=queue[qt]+1;
      qh=(qh+1)%max_queue;
      if (qh == qt  ||  count == max_queue)
        {
        MessageBox(NULL,"Error","MAX_QUEUE exceeded!",MB_OK | MB_APPLMODAL);
        exit(0);
        }
      }
    }
  qt=(qt+1)%max_queue;
  }
(*total_indices)=count;
if (qt != qh)
  return(0);
else
  return(1);
}






	/*******************************************************************
	** Traces each endpoint back to the nearest branch (or other
	** endpoint if there are no branches).  For each trace, the
	** ratio of the trace distance to the straight-line distance
	** is computed, as a rough curvature estimate.  The maximum
	** estimate for the segment is returned.
	*******************************************************************/

void SegmentCurvature(unsigned char	*image,
		int		ROWS,int COLS,
		int		Label,		/* in image, label of segment */
		int		*indices,	/* coded y*COLS+x, points in segment */
		int		count,		/* how many points in segment */
		int		*endpoints,	/* coded y*COLS+x endpoints to trace */
		int		total_ends,	/* number of points to trace */
		int		*branches,	/* coded y*COLS+x, possible stop points */
		int		total_branches,	/* number of branches */
		int		*queue,		/* chunk of memory to use during tracing */
		int		max_queue,	/* size of chunk of memory */
		double	*curvature)	/* returned:  largest curvature in segment */

{
int	qh,qt,r2,c2,x1,y1,x2,y2,e,b,trace_count;
double	dist;

(*curvature)=0;
for (e=0; e<total_ends; e++)
  {
  if (image[endpoints[e]] != Label)
    continue;
  image[endpoints[e]]=4;
  queue[0]=endpoints[e];
  qh=1;
  qt=0;
  trace_count=1;
  x2=y2=-1;
  while (qt != qh)
    {
    for (r2=-1; r2<=1; r2++)
      for (c2=-1; c2<=1; c2++)
        {
        if (r2 == 0  &&  c2 == 0)
          continue;
        if ((queue[qt]/COLS+r2) < 0  ||  (queue[qt]/COLS+r2) >= ROWS  ||
	  	(queue[qt]%COLS+c2) < 0  ||  (queue[qt]%COLS+c2) >= COLS)
          continue;
        if (image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2] != Label)
          continue;
        for (b=0; b<total_branches; b++)
          if (branches[b] == (queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2)
            {
	    y2=queue[qt]/COLS+r2;
	    x2=queue[qt]%COLS+c2;
	    r2=2; qt=qh-1; break;
	    }
        if (b < total_branches)
          break;
        image[(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2]=4;
        trace_count++;
        queue[qh]=(queue[qt]/COLS+r2)*COLS+queue[qt]%COLS+c2;
        qh=(qh+1)%max_queue;
        if (qh == qt)
          {
          MessageBox(NULL,"Error","max_queue exceeded in SegmentCurvature",MB_OK | MB_APPLMODAL);
          exit(0);
          }
        }
    qt=(qt+1)%max_queue;
    }
  for (r2=0; r2<count; r2++)
    if (image[indices[r2]] == 4)
      image[indices[r2]]=Label;
  if (x2 == -1  &&  y2 == -1)
    {		/* no branches -- end to end */
    y2=endpoints[(e+1)%total_ends]/COLS;
    x2=endpoints[(e+1)%total_ends]%COLS;
    }
  if (trace_count < 5)
    continue;
  x1=endpoints[e]%COLS; y1=endpoints[e]/COLS;
  dist=sqrt((double)(SQR(x1-x2)+SQR(y1-y2)));
  if ((double)trace_count/dist > *curvature)
    (*curvature)=(double)trace_count/dist;
  }
}



